@@ -1,7 +1,7 @@
 /*
 	DDS GIMP plugin
 
-	Copyright (C) 2004 Shawn Kirst <skirst@fuse.net>,
+	Copyright (C) 2004-2008 Shawn Kirst <skirst@insightbb.com>,
    with parts (C) 2003 Arne Reuter <homepage@arnereuter.de> where specified.
 
 	This program is free software; you can redistribute it and/or
@@ -32,12 +32,22 @@ typedef struct
    int transindex;
    int color_type;
    int dither;
+   int mipmap_filter;
+   int gamma_correct;
+   float gamma;
    int show_adv_opt;
-} DDSSaveVals;
+} DDSWriteVals;
 
-extern DDSSaveVals ddsvals;
+typedef struct
+{
+   int show_dialog;
+   int mipmaps;
+} DDSReadVals;
 
-extern gint32 read_dds(gchar *);
+extern DDSWriteVals dds_write_vals;
+extern DDSReadVals dds_read_vals;
+
+extern GimpPDBStatusType read_dds(gchar *filename, gint32 *imageID);
 extern GimpPDBStatusType write_dds(gchar *, gint32, gint32);
 
 extern gint interactive_dds;
@@ -48,4 +58,8 @@ extern FILE *errorFile;
 #define LOAD_PROC "file-dds-load"
 #define SAVE_PROC "file-dds-save"
 
+#define DECODE_YCOCG_PROC "color-decode-ycocg"
+#define DECODE_YCOCG_SCALED_PROC "color-decode-ycocg-scaled"
+#define DECODE_ALPHA_EXP_PROC "color-decode-alpha-exp"
+
 #endif
