@@ -1,7 +1,7 @@
 /*
 	DDS GIMP plugin
 
-	Copyright (C) 2004 Shawn Kirst <skirst@fuse.net>,
+	Copyright (C) 2004-2008 Shawn Kirst <skirst@insightbb.com>,
    with parts (C) 2003 Arne Reuter <homepage@arnereuter.de> where specified.
 
 	This program is free software; you can redistribute it and/or
@@ -22,8 +22,11 @@
 
 /*
  * Parts of this code have been generously released in the public domain
- * by 'Ryg'.  The original code can be found (at the time of writing)
- * here:  https://mollyrocket.com/forums/viewtopic.php?t=392
+ * by Fabian 'ryg' Giesen.  The original code can be found (at the time
+ * of writing) here:  http://mollyrocket.com/forums/viewtopic.php?t=392
+ * 
+ * For more information about this code, see the README.dxt file that
+ * came with the source.
  */ 
 
 #include <stdlib.h>
@@ -34,38 +37,51 @@
 #include "dds.h"
 #include "endian.h"
 #include "mipmap.h"
+#include "imath.h"
 
 #include "dxt_tables.h"
 
-#ifndef MIN
-#define MIN(a, b)  ((a) < (b) ? (a) : (b))
-#endif
-#ifndef MAX
-#define MAX(a, b)  ((a) > (b) ? (a) : (b))
-#endif
-
 /* extract 4x4 BGRA block */
-static void extract_block(const unsigned char *src, int w,
-                          unsigned char *block)
+static void extract_block(const unsigned char *src, int x, int y,
+                          int w, int h, unsigned char *block)
 {
-   int i;
+   int i, j;
+   int bw = MIN(w - x, 4);
+   int bh = MIN(h - y, 4);
+   int bx, by;
+   const int rem[] =
+   {
+      0, 0, 0, 0,
+      0, 1, 0, 1,
+      0, 1, 2, 0,
+      0, 1, 2, 3
+   };
+   
    for(i = 0; i < 4; ++i)
    {
-      memcpy(&block[i * 4 * 4], src, 4 * 4);
-      src += w * 4;
+      by = rem[(bh - 1) * 4 + i] + y;
+      for(j = 0; j < 4; ++j)
+      {
+         bx = rem[(bw - 1) * 4 + j] + x;
+         block[(i * 4 * 4) + (j * 4) + 0] =
+            src[(by * (w * 4)) + (bx * 4) + 0];
+         block[(i * 4 * 4) + (j * 4) + 1] =
+            src[(by * (w * 4)) + (bx * 4) + 1];
+         block[(i * 4 * 4) + (j * 4) + 2] =
+            src[(by * (w * 4)) + (bx * 4) + 2];
+         block[(i * 4 * 4) + (j * 4) + 3] =
+            src[(by * (w * 4)) + (bx * 4) + 3];
+      }
    }
 }
 
-static inline int mul8bit(int a, int b)
-{
-   int t = a * b + 128;
-   return((t + (t >> 8)) >> 8);
-}
-
 /* pack BGR8 to RGB565 */
 static inline unsigned short pack_rgb565(const unsigned char *c)
 {
-   return(((c[2] >> 3) << 11) | ((c[1] >> 2) << 5) | (c[0] >> 3));
+   //return(((c[2] >> 3) << 11) | ((c[1] >> 2) << 5) | (c[0] >> 3));
+   return((mul8bit(c[2], 31) << 11) |
+          (mul8bit(c[1], 63) <<  5) |
+          (mul8bit(c[0], 31)      ));
 }
 
 /* unpack RGB565 to BGR */
@@ -82,9 +98,9 @@ static void unpack_rgb565(unsigned char *dst, unsigned short v)
 
 static void lerp_rgb(unsigned char *dst, unsigned char *a, unsigned char *b, int f)
 {
-   dst[0] = a[0] + mul8bit(b[0] - a[0], f);
-   dst[1] = a[1] + mul8bit(b[1] - a[1], f);
-   dst[2] = a[2] + mul8bit(b[2] - a[2], f);
+   dst[0] = blerp(a[0], b[0], f);
+   dst[1] = blerp(a[1], b[1], f);
+   dst[2] = blerp(a[2], b[2], f);
 }
 
 static int color_distance(const unsigned char *c0,
@@ -100,13 +116,17 @@ static int luminance(const unsigned char *c)
    return((c[2] * 54 + c[1] * 182 + c[0] * 20) >> 8);
 }
 
+/* Block dithering function.  Simply dithers a block to 565 RGB.
+ * (Floyd-Steinberg)
+ */
 static void dither_block(unsigned char *dst, const unsigned char *block)
 {
    int err[8], *ep1 = err, *ep2 = err + 4, *tmp;
    int c, y;
    unsigned char *bp, *dp;
    const unsigned char *quant;
-   
+  
+   /* process channels seperately */
    for(c = 0; c < 3; ++c)
    {
       bp = (unsigned char *)block;
@@ -120,18 +140,23 @@ static void dither_block(unsigned char *dst, const unsigned char *block)
       
       for(y = 0; y < 4; ++y)
       {
+         /* pixel 0 */
          dp[ 0] = quant[bp[ 0] + ((3 * ep2[1] + 5 * ep2[0]) >> 4)];
          ep1[0] = bp[ 0] - dp[ 0];
          
+         /* pixel 1 */
          dp[ 4] = quant[bp[ 4] + ((7 * ep1[0] + 3 * ep2[2] + 5 * ep2[1] + ep2[0]) >> 4)];
          ep1[1] = bp[ 4] - dp[ 4];
          
+         /* pixel 2 */
          dp[ 8] = quant[bp[ 8] + ((7 * ep1[1] + 3 * ep2[3] + 5 * ep2[2] + ep2[1]) >> 4)];
          ep1[2] = bp[ 8] - dp[ 8];
          
+         /* pixel 3 */
          dp[12] = quant[bp[12] + ((7 * ep1[2] + 5 * ep2[3] + ep2[2]) >> 4)];
          ep1[3] = bp[12] - dp[12];
          
+         /* advance to next line */
          tmp = ep1;
          ep1 = ep2;
          ep2 = tmp;
@@ -142,6 +167,7 @@ static void dither_block(unsigned char *dst, const unsigned char *block)
    }
 }
 
+/* Color matching function */
 static unsigned int match_colors_block(const unsigned char *block,
                                        unsigned char color[4][3],
                                        int dither)
@@ -166,6 +192,7 @@ static unsigned int match_colors_block(const unsigned char *block,
    
    if(!dither)
    {
+      /* the version without dithering is straight-forward */
       for(i = 15; i >= 0; --i)
       {
          mask <<= 2;
@@ -179,6 +206,7 @@ static unsigned int match_colors_block(const unsigned char *block,
    }
    else
    {
+      /* with floyd-steinberg dithering (see above) */
       int err[8], *ep1 = err, *ep2 = err + 4, *tmp;
       int *dp = dots, y, lmask, step;
       
@@ -190,6 +218,7 @@ static unsigned int match_colors_block(const unsigned char *block,
       
       for(y = 0; y < 4; ++y)
       {
+         /* pixel 0 */
          dot = (dp[0] << 4) + (3 * ep2[1] + 5 * ep2[0]);
          if(dot < halfpt)
             step = (dot < c0pt) ? 1 : 3;
@@ -199,6 +228,7 @@ static unsigned int match_colors_block(const unsigned char *block,
          ep1[0] = dp[0] - stops[step];
          lmask = step;
          
+         /* pixel 1 */
          dot = (dp[1] << 4) + (7 * ep1[0] + 3 * ep2[2] + 5 * ep2[1] + ep2[0]);
          if(dot < halfpt)
             step = (dot < c0pt) ? 1 : 3;
@@ -208,6 +238,7 @@ static unsigned int match_colors_block(const unsigned char *block,
          ep1[1] = dp[1] - stops[step];
          lmask |= step << 2;
 
+         /* pixel 2 */
          dot = (dp[2] << 4) + (7 * ep1[1] + 3 * ep2[3] + 5 * ep2[2] + ep2[1]);
          if(dot < halfpt)
             step = (dot < c0pt) ? 1 : 3;
@@ -217,6 +248,7 @@ static unsigned int match_colors_block(const unsigned char *block,
          ep1[2] = dp[2] - stops[step];
          lmask |= step << 4;
          
+         /* pixel 3 */
          dot = (dp[3] << 4) + (7 * ep1[2] + 5 * ep2[3] + ep2[2]);
          if(dot < halfpt)
             step = (dot < c0pt) ? 1 : 3;
@@ -226,6 +258,7 @@ static unsigned int match_colors_block(const unsigned char *block,
          ep1[3] = dp[3] - stops[step];
          lmask |= step << 6;
 
+         /* advance to next line */
          tmp = ep1;
          ep1 = ep2;
          ep2 = tmp;
@@ -238,6 +271,10 @@ static unsigned int match_colors_block(const unsigned char *block,
    return(mask);
 }
 
+/* Special case color matching for DXT1 color blocks with non-opaque
+ * alpha values.  Simple distance based color matching.  This is my 
+ * little hack, Fabian had no need for DXT1-alpha :)
+ */
 static unsigned int match_colors_block_DXT1alpha(const unsigned char *block,
                                                  unsigned char color[4][3])
 {
@@ -264,6 +301,7 @@ static unsigned int match_colors_block_DXT1alpha(const unsigned char *block,
    return(mask);
 }
 
+/* The color optimization function. (Clever code, part 1) */
 static void optimize_colors_block(const unsigned char *block,
                                   unsigned short *max16, unsigned short *min16)
 {
@@ -277,6 +315,7 @@ static void optimize_colors_block(const unsigned char *block,
    float covf[6], vfr, vfg, vfb, magn;
    float fr, fg, fb;
    
+   /* determine color distribution */
    for(c = 0; c < 3; ++c)
    {
       bp = (unsigned char *)block + c;
@@ -296,6 +335,7 @@ static void optimize_colors_block(const unsigned char *block,
    
    memset(cov, 0, sizeof(cov));
    
+   /* determine covariance matrix */
    for(i = 0; i < 16; ++i)
    {
       b = block[4 * i + 0] - mu[0];
@@ -310,6 +350,7 @@ static void optimize_colors_block(const unsigned char *block,
       cov[5] += b * b;
    }
    
+   /* convert covariance matrix to float, find principal axis via power iter */
    for(i = 0; i < 6; ++i)
       covf[i] = cov[i] / 255.0f;
    
@@ -334,7 +375,7 @@ static void optimize_colors_block(const unsigned char *block,
    
    magn = MAX(MAX(vfr, vfg), vfb);
    
-   if(magn < 4.0)
+   if(magn < 4.0) /* too small, default to luminance */
    {
       r = 148;
       g = 300;
@@ -348,6 +389,7 @@ static void optimize_colors_block(const unsigned char *block,
       b = (int)(vfb * magn);
    }
    
+   /* pick colors at extreme points */
    mnd =  0x7fffffff;
    mxd = -0x7fffffff;
    
@@ -367,16 +409,23 @@ static void optimize_colors_block(const unsigned char *block,
       }
    }
 
+   /* reduce to 16-bit colors */
    *max16 = pack_rgb565(mxc);
    *min16 = pack_rgb565(mnc);
 }
 
+/* The refinement function (Clever code, part 2)
+ * Tries to optimize colors to suit block contents better.
+ * (By solving a least squares system via normal equations + Cramer's rule)
+ */
 static int refine_block(const unsigned char *block,
                         unsigned short *max16, unsigned short *min16,
                         unsigned int mask)
 {
    static const int w1tab[4] = {3, 0, 2, 1};
    static const int prods[4] = {0x090000, 0x000900, 0x040102, 0x010402};
+   /* ^ Some magic to save a lot of multiplies in the accumulating loop... */
+   
    int akku = 0;
    int At1_r, At1_g, At1_b;
    int At2_r, At2_g, At2_b;
@@ -386,7 +435,29 @@ static int refine_block(const unsigned char *block,
    float frb, fg;
    unsigned short v, oldmin, oldmax;
    int s;
-   
+
+   oldmin = *min16;
+   oldmax = *max16;
+   if((mask ^ (mask << 2)) < 4) /* all pixels have the same index */
+   {
+      /* degenerate system, use optimal single-color match for average color */
+      r = g = b = 8;
+      for(i = 0; i < 16; ++i)
+      {
+         r += block[4 * i + 2];
+         g += block[4 * i + 1];
+         b += block[4 * i + 0];
+      }
+      
+      r >>= 4;
+      g >>= 4;
+      b >>= 4;
+      
+      *max16 = (omatch5[r][0] << 11) | (omatch6[g][0] << 5) | omatch5[b][0];
+      *min16 = (omatch5[r][1] << 11) | (omatch6[g][1] << 5) | omatch5[b][1];
+      return(*min16 != oldmin || *max16 != oldmax);
+   }
+
    At1_r = At1_g = At1_b = 0;
    At2_r = At2_g = At2_b = 0;
    
@@ -410,20 +481,16 @@ static int refine_block(const unsigned char *block,
    At2_r = 3 * At2_r - At1_r;
    At2_g = 3 * At2_g - At1_g;
    At2_b = 3 * At2_b - At1_b;
-   
+
+   /* extract solutions and decide solvability */
    xx = akku >> 16;
    yy = (akku >> 8) & 0xff;
    xy = (akku >> 0) & 0xff;
    
-   if(!yy || !xx || xx * yy == xy * xy)
-      return(0);
-   
    frb = 3.0f * 31.0f / 255.0f / (xx * yy - xy * xy);
    fg = frb * 63.0f / 31.0f;
    
-   oldmin = *min16;
-   oldmax = *max16;
-
+   /* solve */
    s = (int)((At1_r * yy - At2_r * xy) * frb + 0.5f);
    if(s < 0) s = 0;
    if(s > 31) s = 31;
@@ -451,10 +518,11 @@ static int refine_block(const unsigned char *block,
    if(s > 31) s = 31;
    v |= s;
    *min16 = v;
-   
+
    return(oldmin != *min16 || oldmax != *max16);
 }
 
+/* Find min/max colors by distance */
 static void get_min_max_colors_distance(const unsigned char *block,
                                         unsigned short *max16,
                                         unsigned short *min16)
@@ -480,6 +548,7 @@ static void get_min_max_colors_distance(const unsigned char *block,
    *min16 = MIN(c0, c1);
 }
 
+/* Find min-max colors by luminance */
 static void get_min_max_colors_luminance(const unsigned char *block,
                                          unsigned short *max16,
                                          unsigned short *min16)
@@ -508,6 +577,7 @@ static void get_min_max_colors_luminance(const unsigned char *block,
 
 #define INSET_SHIFT  4
 
+/* Find min-max colors using the inset bounding box method */
 static void get_min_max_colors_inset_bbox(const unsigned char *block,
                                           unsigned short *max16,
                                           unsigned short *min16)
@@ -686,17 +756,21 @@ static void encode_color_block(unsigned char *dst,
    unsigned int v, mn, mx, mask;
    int i, block_has_alpha = 0;
 
+   /* find min/max colors, determine if alpha values present in block
+    * (for DXT1-alpha)
+    */
    mn = mx = GETL32(block);
    for(i = 0; i < 16; ++i)
    {
       block_has_alpha = block_has_alpha || (block[4 * i + 3] < 255);
       v = GETL32(&block[4 * i]);
-      if(v > mx) mx = v;
-      if(v < mn) mn = v;
+      mx = MAX(mx, v);
+      mn = MIN(mn, v);
    }
    
-   if(mn != mx)
+   if(mn != mx) /* block is not a solid color, continue with compression */
    {
+      /* compute dithered block for PCA if desired */
       if(dither)
          dither_block(dblock, block);
       
@@ -712,6 +786,7 @@ static void encode_color_block(unsigned char *dst,
             get_min_max_colors_inset_bbox(dither ? dblock : block, &max16, &min16);
             break;
          default:
+            /* pca + map along principal axis */
             optimize_colors_block(dither ? dblock : block, &max16, &min16);
             if(max16 != min16)
             {
@@ -720,7 +795,8 @@ static void encode_color_block(unsigned char *dst,
             }
             else
                mask = 0;
-      
+         
+            /* refine */
             refine_block(dither ? dblock : block, &max16, &min16, mask);
             break;
       }
@@ -733,7 +809,7 @@ static void encode_color_block(unsigned char *dst,
       else
          mask = 0;
    }
-   else
+   else /* constant color */
    {
       mask = 0xaaaaaaaa;
       max16 = (omatch5[block[2]][0] << 11) |
@@ -744,6 +820,7 @@ static void encode_color_block(unsigned char *dst,
               (omatch5[block[0]][1]      );
    }
    
+   /* HACK! for DXT1 blocks which have non-opaque pixels */
    if(dxt1_alpha && block_has_alpha)
    {
       if(max16 > min16)
@@ -765,6 +842,7 @@ static void encode_color_block(unsigned char *dst,
    PUTL32(&dst[4], mask);
 }
 
+/* write DXT3 alpha block */
 static void encode_alpha_block_DXT3(unsigned char *dst,
                                     const unsigned char *block)
 {
@@ -780,15 +858,19 @@ static void encode_alpha_block_DXT3(unsigned char *dst,
    }
 }
 
+/* Write DXT5 alpha block */
 static void encode_alpha_block_DXT5(unsigned char *dst,
-                                    const unsigned char *block)
+                                    const unsigned char *block,
+                                    const int offset)
 {
    int i, v, mn, mx;
    int dist, bias, dist2, dist4, bits, mask;
    int a, idx, t;
    
+   block += offset;
    block += 3;
    
+   /* find min/max alpha pair */
    mn = mx = block[0];
    for(i = 0; i < 16; ++i)
    {
@@ -797,9 +879,11 @@ static void encode_alpha_block_DXT5(unsigned char *dst,
       if(v < mn) mn = v;
    }
    
+   /* encode them */
    *dst++ = mx;
    *dst++ = mn;
    
+   /* determine bias and emit indices */
    dist = mx - mn;
    bias = mn * 7 - (dist >> 1);
    dist4 = dist * 4;
@@ -811,6 +895,7 @@ static void encode_alpha_block_DXT5(unsigned char *dst,
    {
       a = block[4 * i] * 7 - bias;
       
+      /* select index (hooray for bit magic) */
       t = (dist4 - a) >> 31; idx =  t & 4; a -= dist4 & t;
       t = (dist2 - a) >> 31; idx += t & 2; a -= dist2 & t;
       t = (dist  - a) >> 31; idx += t & 1;
@@ -818,6 +903,7 @@ static void encode_alpha_block_DXT5(unsigned char *dst,
       idx = -idx & 7;
       idx ^= (2 > idx);
       
+      /* write index */
       mask |= idx << bits;
       if((bits += 3) >= 8)
       {
@@ -834,11 +920,11 @@ static void compress_DXT1(unsigned char *dst, const unsigned char *src,
    unsigned char block[64];
    int x, y;
    
-   for(y = 0; y < h; y += 4, src += w * 4 * 4)
+   for(y = 0; y < h; y += 4)
    {
       for(x = 0; x < w; x += 4)
       {
-         extract_block(src + x * 4, w, block);
+         extract_block(src, x, y, w, h, block); 
          encode_color_block(dst, block, type, dither, alpha);
          dst += 8;
       }
@@ -851,11 +937,11 @@ static void compress_DXT3(unsigned char *dst, const unsigned char *src,
    unsigned char block[64];
    int x, y;
    
-   for(y = 0; y < h; y += 4, src += w * 4 * 4)
+   for(y = 0; y < h; y += 4)
    {
       for(x = 0; x < w; x += 4)
       {
-         extract_block(src + x * 4, w, block);
+         extract_block(src, x, y, w, h, block);
          encode_alpha_block_DXT3(dst, block);
          encode_color_block(dst + 8, block, type, dither, 0);
          dst += 16;
@@ -869,12 +955,12 @@ static void compress_DXT5(unsigned char *dst, const unsigned char *src,
    unsigned char block[64];
    int x, y;
    
-   for(y = 0; y < h; y += 4, src += w * 4 * 4)
+   for(y = 0; y < h; y += 4)
    {
       for(x = 0; x < w; x += 4)
       {
-         extract_block(src + x * 4, w, block);
-         encode_alpha_block_DXT5(dst, block);
+         extract_block(src, x, y, w, h, block);
+         encode_alpha_block_DXT5(dst, block, 0);
          encode_color_block(dst + 8, block, type, dither, 0);
          dst += 16;
       }
@@ -887,12 +973,12 @@ static void compress_BC4(unsigned char *dst, const unsigned char *src,
    unsigned char block[64];
    int x, y;
    
-   for(y = 0; y < h; y += 4, src += w * 4 * 4)
+   for(y = 0; y < h; y += 4)
    {
       for(x = 0; x < w; x += 4)
       {
-         extract_block(src + x * 4, w, block);
-         encode_alpha_block_DXT5(dst, block - 1);
+         extract_block(src, x, y, w, h, block);
+         encode_alpha_block_DXT5(dst, block, -1);
          dst += 8;
       }
    }
@@ -904,13 +990,13 @@ static void compress_BC5(unsigned char *dst, const unsigned char *src,
    unsigned char block[64];
    int x, y;
    
-   for(y = 0; y < h; y += 4, src += w * 4 * 4)
+   for(y = 0; y < h; y += 4)
    {
       for(x = 0; x < w; x += 4)
       {
-         extract_block(src + x * 4, w, block);
-         encode_alpha_block_DXT5(dst, block - 2);
-         encode_alpha_block_DXT5(dst + 8, block - 1);
+         extract_block(src, x, y, w, h, block);
+         encode_alpha_block_DXT5(dst, block, -2);
+         encode_alpha_block_DXT5(dst + 8, block, -1);
          dst += 16;
       }
    }
@@ -927,13 +1013,13 @@ static void compress_YCoCg(unsigned char *dst, const unsigned char *src,
    int x0, x1, x2;
    int x, y, i;
    
-   for(y = 0; y < h; y += 4, src += w * 4 * 4)
+   for(y = 0; y < h; y += 4)
    {
       for(x = 0; x < w; x += 4)
       {
-         extract_block(src + x * 4, w, block);
+         extract_block(src, x, y, w, h, block);
          
-         encode_alpha_block_DXT5(dst, block);
+         encode_alpha_block_DXT5(dst, block, 0);
          
          maxcolor = &colors[0][0];
          mincolor = &colors[1][0];
@@ -969,7 +1055,7 @@ static void compress_YCoCg(unsigned char *dst, const unsigned char *src,
             x2 = b0 & b4;
 
             mask <<= 2;
-            mask |= (x1 | ((x0 | x1) << 1));
+            mask |= (x2 | ((x0 | x1) << 1));
          }
          
          PUTL16(&dst[ 8], pack_rgb565(maxcolor));
@@ -983,7 +1069,8 @@ static void compress_YCoCg(unsigned char *dst, const unsigned char *src,
 
 int dxt_compress(unsigned char *dst, unsigned char *src, int format,
                  unsigned int width, unsigned int height, int bpp,
-                 int mipmaps, int type, int dither)
+                 int mipmaps, int type, int dither, int filter,
+                 int gamma_correct, float gamma)
 {
    int i, size, w, h;
    unsigned int offset;
@@ -992,13 +1079,17 @@ int dxt_compress(unsigned char *dst, unsigned char *src, int format,
    unsigned char *tmp2, *s;
    int dxt1_alpha = 0;
    
-   if(!(IS_POT(width) && IS_POT(height)))
+   if(!(IS_MUL4(width) && IS_MUL4(height)))
+      return(0);
+   
+   if((mipmaps > 1) && !(IS_POW2(width) && IS_POW2(height)))
       return(0);
    
    size = get_mipmapped_size(width, height, bpp, 0, mipmaps,
                              DDS_COMPRESS_NONE);
    tmp = g_malloc(size);
-   generate_mipmaps(tmp, src, width, height, bpp, 0, mipmaps);
+   generate_mipmaps(tmp, src, width, height, bpp, 0, mipmaps, filter,
+                    gamma_correct, gamma);
    
    if(bpp == 4 && format == DDS_COMPRESS_BC1)
       dxt1_alpha = 1;
@@ -1095,8 +1186,8 @@ int dxt_compress(unsigned char *dst, unsigned char *src, int format,
       }
       s += (w * h * bpp);
       offset += get_mipmapped_size(w, h, 0, 0, 1, format);
-      if(w > 1) w >>= 1;
-      if(h > 1) h >>= 1;
+      w = MAX(1, w >> 1);
+      h = MAX(1, h >> 1);
    }
 
    g_free(tmp);
@@ -1211,7 +1302,7 @@ int dxt_decompress(unsigned char *dst, unsigned char *src, int format,
    unsigned char *d, *s;
    unsigned int x, y, sx, sy;
    
-   if(!(IS_POT(width) && IS_POT(height)))
+   if(!(IS_MUL4(width) && IS_MUL4(height)))
       return(0);
    
    sx = (width  < 4) ? width  : 4;
